// C Source File
// Created 08.07.2005
// Copyright (C) 2005 iga Mirai
//QMC V7 Quine-McCluskey boolean expression minimization

// V7 08.07.2005
//  Dodan je opcijski argument nrand
//   e vnese nrand>0, potem v drugem delu ne bo iskal najmanjega pokritja po vrsti
//   ez vse pprod ampak bo pprod generiral nakljuno
//
// *pprod dejansko oznauje particijski produkt iz zgodbe
//  o minimizaciji stanj asinh. avtomatov in je ostal
//  iz zgodovinskih razlogov*

// V6 08.07.2005
//  V drugem delu sedaj lahko pritisne 'ESC' in dobi trenutno najbolji rezultat
//   ki ni nujno minimalno pokritje in niso nujno vse mone reitve te doline
//  Informacja ki jo prikazuje je sedaj: t. produktnih lenov in t. reitev

// V5 07.07.2005
//  SizeableShortArray1 sedaj kljuuje zaito v primeru da malloc vrne NULL
//   in se varno vrne brez puanja spomina

// V4 06.07.2005
//  reduce in sort so sedaj funkcije
//  veliko bolj stabilna, ker se prime tabeli e v prvem delu kraja
//  v drugem delu se tudi za silo preverja mono t. kombinacij KK da se tevec ne obrne
//  dodal e preverjanje, e je lista mintermov prazna

// V3 05.07.2005
//  nekoliko kraja, ker: SizeableArray -> SizeableShortArray
//  deluje stabilno do SEQ(n,n,0,30)
//  potem zmanjka spomina v prvem delu
//  drugega dela pa ne bi akal za ve kot 0..27

#define USE_TI89              // Produce .89z File
#define RETURN_VALUE          // Return Pushed Expression
#define OPTIMIZE_ROM_CALLS    // Use ROM Call Optimization
// #define SAVE_SCREEN        // Save/Restore LCD Contents

//>>>> UNCOMMENT THE FOLLOWING LINE to get return value stored in 'result' <<<<
//#define RETURN_VALUE result	  // Store Pushed Expression into result variable


//Includes
//----------------------------------------------------------------------------------
#include <tigcclib.h>         // Include All Header Files
#include "PomozneF.h"
#include "SizeableShortArray1.h"
#include "QMC.h"


// Globals
//----------------------------------------------------------------------------------
#define LQ 64 //za koliko se liste poveujejo
char StrAbout[]="Quine-McCluskey method implemented by Ziga Mirai, July 2005";
char EArgs[]="E: usage QMC(Minterms_list, Dontcares_list, [nRand]) !";
char EExcessIter[]="E: More than 8388607 iterations required!";
char ENoMinterms[]="E: At least one minterm required!";


// Main
//----------------------------------------------------------------------------------
void _main(void)
{
	USE_LISTS;
	//e ni argumenta izpii About
	//e niso pravi pa error in konaj
	unsigned short argc=remaining_element_count(top_estack);
	if(argc==0) {
		top_estack=PushString(top_estack,StrAbout);
		return;
	} else if(argc!=2 && argc!=3) ETxtExit(EArgs);
	ESI ep=top_estack;
	ESI epd=next_expression_index(top_estack);
	if(GetArgType(ep)!=LIST_TAG || GetArgType(epd)!=LIST_TAG) ETxtExit(EArgs);
	
	//e imamo opcijski argument nRand
	long nRand=0;
	if(argc==3) {
		ESI epr=next_expression_index(epd);
		if(GetArgType(epr)!=POSINT_TAG) ETxtExit(EArgs);
		nRand=GetIntArg(epr);
		randomize();
	}

	//naloi listi mi in d
	short miLen=remaining_element_count(ep-1);
	short dLen=remaining_element_count(epd-1);
	CREATE_LIST(mi,miLen,LQ);
	CREATE_LIST(d,dLen,LQ);
	ep=LoadShortIntList(mi.a, ep);				//preberi minterms list
	top_estack=LoadShortIntList(d.a, ep);	//preberi dontcares list

	//e ni mintermov
	if(miLen==0) {
		top_estack=PushString(top_estack, ENoMinterms);
		DELETE_LIST(d);
		DELETE_LIST(mi);
		return;
	}

	//poii prime implicants oz. minimiziraj dolino produktov
	//mi in d skupaj in zaetne maske m
	AUGMENT_LIST(d,mi);
	CREATE_LIST(m,d.N,LQ);
	//primes in primes masks
	CREATE_LIST(prl,0,LQ);
	CREATE_LIST(prm,0,LQ);
	//implicants in imp. masks
	CREATE_LIST(newd,0,LQ);
	CREATE_LIST(newm,0,LQ);
	//combines w. someone flags
	CREATE_LIST(t,0,LQ);

//+
	char dbgstr[128];
	void *kbd=kbd_queue();
	unsigned short key;
//-

	//pri = prl.N
	long i,j;
	unsigned short i1n;
	while(1) {
		//implicants in imp. masks
		DELETE_LIST(newd);	//{}->newd
		DELETE_LIST(newm);	//{}->newm
		//combines w. someone flags
		DELETE_LIST(t);		//newlist(dd)->t
		W_L(t, d.N-1, 0);	//dd = d.N
		//ni = newd.N
		short dai,mai;
		for(i=0;i<d.N-1;i++) {
//+
			if(d.N>100 && (i&0xf)==0) {
				sprintf(dbgstr,"d%li:%li  ",i,d.N);
				DrawStrXY(0,74,dbgstr,A_REPLACE);
				//read keys
				if(!OSdequeue(&key,kbd)) {
					//e pritisne 'ON' prekini izvajanje
					if(key==KEY_ON) {
						push_longint(0); //da vrne Error Break namesto Done ko top_estack--
						DeleteAllLists();
						return;
					}
				}
			}
//-
			dai=d.a[i];
			mai=m.a[i];
			for(j=i+1;j<d.N;j++) {
				if(mai==m.a[j]) {
					i1n=Is1Near((unsigned)dai,(unsigned)d.a[j]);
					if(i1n!=0) {
						t.a[i]=1;
						t.a[j]=1;
						PUSH_L(newd, dai|i1n);
						PUSH_L(newm, mai|i1n);
					}
				}
			}
			if(t.a[i]==0) {
				PUSH_L(prl, dai);
				PUSH_L(prm, mai);
			}
		}
		if(t.a[i]==0) {
			PUSH_L(prl, d.a[i]);
			PUSH_L(prm, m.a[i]);
		}
		if(newd.N==0)
			break;
		else {
			//tudi na vmesnih korakih reduciraj listo pirmeov ...
			newm.N=newd.N=LReduceShort2(newd.a, newm.a, newd.N);
			//predno gredo v novo iteracijo
			DELETE_LIST(d); AUGMENT_LIST(d,newd);	//newd->d
			DELETE_LIST(m); AUGMENT_LIST(m,newm);	//newm->m
		}
	}

	DELETE_LIST(t);	//tagov ne rabim ve
	DELETE_LIST(d);	//d ne rabim ve
	DELETE_LIST(m);	//m ne rabim ve

	//odstrani enake prime.i. iz prl in prm
	prm.N=prl.N=LReduceShort2(prl.a, prm.a, prl.N);

	//pretej s koliko prime implikanti je pokrit vsak minterm
	DELETE_LIST(newm); W_L(newm, miLen-1, 0);	//newlist(dim(mi))->newm
	DELETE_LIST(newd);	//{}->newd
	//k teje newd
	for(i=0;i<miLen;i++) {
		for(j=0;j<prl.N;j++) {
			if((mi.a[i]|prm.a[j])==prl.a[j]) {
				PUSH_L(newd, j);
				W_L(newm, i, newm.a[i]+1);
			}
		}
	}
	//newm[]=tevilo primeov na minterm
	//newd[]=indeksi primeov za vsak minterm
	DELETE_LIST(mi);	//mi ne rabim ve



	//Generiranje produktnih lenov particijskega polinoma eg. (p1+p4+p5)*(p1+p2)...

	//ri[i]=indeks prima na i-tem mestu
	short ri[miLen]; for(i=0;i<miLen;i++) ri[i]=0;
	//of[i]=kazalec na mesto v newd, ki pripada, i-temu mintermu
	short of[miLen]; of[0]=0; for(i=1;i<miLen;i++) of[i]=newm.a[i-1]+of[i-1];
	//KK=vseh monih kombinacij za particijski produkt
	long KK=1; for(i=0;i<miLen && KK<0x007fffffL;i++) KK*=newm.a[i];
	//e generiramo nakljune reitve, uporabnik doloi t. iteracij
	if(nRand) KK=nRand;
	//da se ne bi tevec obrnil in da ne bi 4194303 iteracij akal
	if(KK>=0x007fffffL) {
		top_estack=PushString(top_estack, EExcessIter);
		DeleteAllLists();
		return;
	}
	//pprod[] = primi (njihovi indeksi) posameznega particijskega produkta eg. p1*p1*p3*p4
	CREATE_LIST(pprod,miLen,LQ);

	//zbiranje najkrajih pprod (vsak je najprej dolg miLen vendar se p1*p1*p1...=p1 skraja)
	//pplist sem se zbirajo enako dolgi najkraji pprodi
	CREATE_LIST(pplist,0,LQ);	//{}->pplist
	short pplen=miLen;	//dolina trenutno najkrajega pprod

//+
	short nCovers=1;
//-
	long k;
	for(k=0;k<KK;k++) {	//po vseh monih pprod
		//sestavi en pprod
		if(nRand==0) {
			//zaporedno generiranje pprod
			for(i=0;i<miLen;i++)
				pprod.a[i]=newd.a[of[i]+ri[i]];
			//tevec
			for(i=0;i<miLen;i++) {
				ri[i]++;
				if(ri[i]<newm.a[i]) break;
				ri[i]=0;
			}
		} else {
			//nakljuno generiranje pprod
			for(i=0;i<miLen;i++)
				pprod.a[i]=newd.a[of[i]+random(newm.a[i])];
		}

		//enake produkte krajat p1*p1=p1
		//lreduce(pprod)->pprod
		pprod.N=LReduceShort1(pprod.a, miLen);

		//Uredi pprod za kasneje metanje ven enakih pprod
		//lsorta(pprod)->pprod
		InsSortShort1(pprod.a, pprod.N);

		//zbiraj najkraje pprod v pplist
		if(pprod.N==pplen) {
			//e si sledijo enaki jih ne dodajaj
			if(memcmp(&pplist.a[pplist.N-pplen], pprod.a, pplen*sizeof(short))!=0 || pplist.N==0) {
				AUGMENT_LIST(pplist,pprod);
			}
			nCovers=pplist.N/pplen;
		} else if(pprod.N<pplen) {
			DELETE_LIST(pplist);				//pprod->pplist
			AUGMENT_LIST(pplist,pprod);	//pprod->pplist
			pplen=pprod.N;
			nCovers=pplist.N/pplen;
		}
//+
		if(KK>127 && ((k&0x7f)==0)) {
			//prikazuje k trenutno kombinacijo : KK vseh kombinacij
			// P trenutno najmanje t. produktov v disjunkciji
			// c trenutno tevilo najdenih pokritij enake doline
			//   (po koncu je lahko manje ker se enaka pokritja zmeejo ven)
			sprintf(dbgstr,"k%li/%li P%i n%i  ",k,KK,pplen,nCovers);
			DrawStrXY(0,74,dbgstr,A_REPLACE);
			//read keys
			if(!OSdequeue(&key,kbd)) {
				if(key==KEY_ON) {
					//e pritisne 'ON' prekini izvajanje
					push_longint(0);
					DeleteAllLists();
					return;
				} else if(key==KEY_ESC) {
					//e pritisne 'ESC' vrni trenutno najmanje pokritje
					break;
				}
			}
		}
//-
	}


	//Vrzi ven enake pprod iz pplist
	//lreducen(pplist, pplen)->pplist
	pplist.N=LReduceShortSubL1(pplist.a, pplist.N, pplen);

	//izgleda da pred matrikami ne sme biti zaetni END_TAG
	//pred listami pa je lahko, vendar ni nujen
	top_estack--;
	//IZPIS: tevilke primeov iz pplist v matirko primeov in mask
	push_quantum(END_TAG);
	short primes[pplen],masks[pplen];
	for(i=0;i<pplist.N;i+=pplen) {
		for(j=0;j<pplen;j++) {
			masks[j]=prm.a[pplist.a[i+j]];
			primes[j]=prl.a[pplist.a[i+j]];
		}
		top_estack=PushShortIntList(top_estack, masks, pplen);
		top_estack=PushShortIntList(top_estack, primes, pplen);
	}
	push_quantum(LIST_TAG);

	DeleteAllLists();
}
