//Sizeable array
//Copyright (C) 2005 Ziga Mirai
//2005.07.07

// namenjen je QMC
// vkljuuje varno alokacijo spomina

#ifndef SZ_ARRAYH
#define SZ_ARRAYH

//---------------------------- Sizeable Array ------------------------------

#define MAX_N_LISTS 20

typedef struct {
	short *a;
	long N;
	long avail;
	long lq;
} TShortList;

//safety pointers to created lists
TShortList *pSLStruct[MAX_N_LISTS];
int nListsCreated;

//inicializira polje kazalcev na kreirane liste
// obvezno vkljuiti na zaetek main funkcije!
#define USE_LISTS nListsCreated=0

//Sprosti spomin vseh kreiranih polj
void DeleteAllLists() {
	int i;
	for(i=0;i<nListsCreated;i++) {
		if(pSLStruct[i]->a!=NULL) {
			free(pSLStruct[i]->a);
			pSLStruct[i]->N=0;
			pSLStruct[i]->avail=0;
			pSLStruct[i]->a=NULL;
		}
	}
}

//Not enough emory or too many lists created safety return
void EMemoryExit() {
	DeleteAllLists();
	EatArgs();
	push_quantum(END_TAG);
	char rs[40];
	sprintf(rs,"E: Not enough emory (%02i)!",nListsCreated);
	top_estack=PushString(top_estack,rs);
	exit(0);
}

//CREATE_SA
// kreira sizeable array strukturo z imenom name
// zapolnjena je z nilami
// klie jo samo enkrat na zaetku
// in vsebuje:
// pub a = polje izbranega tipa atype
// pub N = hrani dolino polja (indeks za najviji vpisani element)
// prot avail = razpololjiva dolina polja za direktno pisanje
// prot lq = dolina za katero se podaljuje polje
#define CREATE_LIST(name, len, len_quant)\
TShortList name={ NULL , len, len, len_quant};\
if(nListsCreated>=MAX_N_LISTS) EMemoryExit();\
pSLStruct[nListsCreated]=&name; nListsCreated++;\
if((len)>0) {\
	name.a=(short *) malloc(len*2);\
	if(name.a==NULL) EMemoryExit();\
	memset(name.a, 0, (len)*2);\
}

//DELETE_LIST
// skraja dolino polja na 0 in sprosti pomnilnik
// lahko se klie kadarkoli, obvezno pa na koncu
#define DELETE_LIST(name) if(name.avail>0) free(name.a);\
name.N=name.avail=0; name.a=NULL;

//W_L
// vpise en element polja
// pri tem povea N e je treba
// in rezervira nov blok spomina, e piemo izven rezervirnega
// dodan blok je popisan z nilami
#define W_L(name,i,e)\
if((i)>=name.avail) ExtendList(&name,i);\
name.a[i]=e;\
if((i)>=name.N) name.N=(i)+1

//PUSH_L
// doda en element na vrh
#define PUSH_L(name,e) W_L(name,name.N,e);

//Extend_LIST
// uporabnik je ne rabi
// je privatna, rezervira nov kos pomnilnika in prepie vsebino
// ko prekorai obstojeo dolino pri vpisovanju
void ExtendList(TShortList *list, long newLen) {
	long newavail=newLen+list->lq;
	if(list->a==NULL)
		list->a=malloc(newavail*2);
	else
		list->a=realloc(list->a, newavail*2);
	if(list->a==NULL) EMemoryExit();
	memset(list->a + list->avail, 0, (newavail - list->avail)*2 );
	list->avail=newavail;
}

//AUGMENT_LIST
//listo name1 podalja z vsebino liste name2
#define AUGMENT_LIST(name1,name2)\
if(name2.N>0) {\
	long oldN=name1.N;\
	W_L(name1, name1.N+name2.N-1,0);\
	memcpy(name1.a+oldN, name2.a, name2.N*2);\
}


//---------------------------^ Sizeable Array ^-----------------------------

#endif
