// C Source File
// Created 08.07.2005
// Copyright (C) 2005 iga Mirai
//QMC V7 Quine-McCluskey boolean expression minimization

// V7 06.04.2006
//  comments translation

// V7 08.07.2005
//  Optional argument nrand added
//   if you choose nrand>0, then, in the second part, it won't search for the
//   min. cover over all pprod but instead, pprod will be random generated
//
// (note: pprod actually stands for partitional product from the story
//  of the minimisation of the asynchronous automata state reduction, and
//  remains here for the historical reasons;
//  I found out later, that it was called Petrick's method when used with
//  Quine-McCluskey - http://en.wikipedia.org/wiki/Petrick%27s_Method )

// V6 08.07.2005
// Now you can press 'ESC' in the second part and get the current best result
//  that is not necessarily the minimal cover and it is not all the possible
//  solutions of this length
// Information that is now displayed is: number of product terms and number of
//  solutions

// V5 07.07.2005
//  SizeableShortArray1 now includes safety mechanism in case malloc returns NULL
//   and safely returns without leaking memory

// V4 06.07.2005
//  reduce and sort are now functions
//  much more stable, because of the primes list size reduction in the first part
//  in the second part possible number of combinations KK is checked
//   to prevent counter overflow
//  also added: checking if the minterms list is empty

// V3 05.07.2005
//  somewhat shorter, because of: SizeableArray -> SizeableShortArray
//  stable to SEQ(n,n,0,30), then: insufficient memory in the first part
//  but I wouldn't wait the second part for more than 0..27

#define USE_TI89              // Produce .89z File
#define RETURN_VALUE          // Return Pushed Expression
#define OPTIMIZE_ROM_CALLS    // Use ROM Call Optimization
// #define SAVE_SCREEN        // Save/Restore LCD Contents

//>>>> UNCOMMENT THE FOLLOWING LINE to get return value stored in 'result' <<<<
//#define RETURN_VALUE result	  // Store Pushed Expression into result variable


//Includes
//----------------------------------------------------------------------------------
#include <tigcclib.h>         // Include All Header Files
#include "HelperF.h"
#include "SizeableShortArray1.h"
#include "QMC.h"


// Globals
//----------------------------------------------------------------------------------
#define LQ 64 //length quantum by wich lists are being estended
char StrAbout[]="Quine-McCluskey method implemented by Ziga Mirai, July 2005";
char EArgs[]="E: usage QMC(Minterms_list, Dontcares_list, [nRand]) !";
char EExcessIter[]="E: More than 8388607 iterations required!";
char ENoMinterms[]="E: At least one minterm required!";


// Main
//----------------------------------------------------------------------------------
void _main(void)
{
	USE_LISTS;
	//if there's no argument print About
	//if there are wrong arguments push error message and exit
	unsigned short argc=remaining_element_count(top_estack);
	if(argc==0) {
		top_estack=PushString(top_estack,StrAbout);
		return;
	} else if(argc!=2 && argc!=3) ETxtExit(EArgs);
	ESI ep=top_estack;
	ESI epd=next_expression_index(top_estack);
	if(GetArgType(ep)!=LIST_TAG || GetArgType(epd)!=LIST_TAG) ETxtExit(EArgs);

	//if optional argument nRand is included
	long nRand=0;
	if(argc==3) {
		ESI epr=next_expression_index(epd);
		if(GetArgType(epr)!=POSINT_TAG) ETxtExit(EArgs);
		nRand=GetIntArg(epr);
		randomize();
	}

	//load lists mi and d
	short miLen=remaining_element_count(ep-1);
	short dLen=remaining_element_count(epd-1);
	CREATE_LIST(mi,miLen,LQ);
	CREATE_LIST(d,dLen,LQ);
	ep=LoadShortIntList(mi.a, ep);			//load minterms list
	top_estack=LoadShortIntList(d.a, ep);	//load dontcares list

	//if there are no minterms
	if(miLen==0) {
		top_estack=PushString(top_estack, ENoMinterms);
		DELETE_LIST(d);
		DELETE_LIST(mi);
		return;
	}

	//find prime implicants (i.e. minimize products length)
	//joined mi and d, initial masks m
	AUGMENT_LIST(d,mi);
	CREATE_LIST(m,d.N,LQ);
	//primes and primes masks
	CREATE_LIST(prl,0,LQ);
	CREATE_LIST(prm,0,LQ);
	//implicants and implicants masks
	CREATE_LIST(newd,0,LQ);
	CREATE_LIST(newm,0,LQ);
	//combines_with_someone flags (tags)
	CREATE_LIST(t,0,LQ);

//+ [ + - mark later insterted debug/user_exit section ]
	char dbgstr[128];
	void *kbd=kbd_queue();
	unsigned short key;
//-

	//pri = prl.N
	long i,j;
	unsigned short i1n;
	while(1) {
		//implicants and implicants masks
		DELETE_LIST(newd);	//{}->newd
		DELETE_LIST(newm);	//{}->newm
		//combines_w._someone tags
		DELETE_LIST(t);		//newlist(dd)->t
		W_L(t, d.N-1, 0);	//dd = d.N
		//ni = newd.N
		short dai,mai;
		for(i=0;i<d.N-1;i++) {
//+
			if(d.N>100 && (i&0xf)==0) {
				sprintf(dbgstr,"d%li:%li  ",i,d.N);
				DrawStrXY(0,74,dbgstr,A_REPLACE);
				//read keys
				if(!OSdequeue(&key,kbd)) {
					//end execution if 'ON' key is pressed
					if(key==KEY_ON) {
						push_longint(0); //to return Error Break instead of Done when top_estack--
						DeleteAllLists();
						return;
					}
				}
			}
//-
			dai=d.a[i];
			mai=m.a[i];
			for(j=i+1;j<d.N;j++) {
				if(mai==m.a[j]) {
					i1n=Is1Near((unsigned)dai,(unsigned)d.a[j]);
					if(i1n!=0) {
						t.a[i]=1;
						t.a[j]=1;
						PUSH_L(newd, dai|i1n);
						PUSH_L(newm, mai|i1n);
					}
				}
			}
			if(t.a[i]==0) {
				PUSH_L(prl, dai);
				PUSH_L(prm, mai);
			}
		}
		if(t.a[i]==0) {
			PUSH_L(prl, d.a[i]);
			PUSH_L(prm, m.a[i]);
		}
		if(newd.N==0)
			break;
		else {
			//reduce primes list ...
			newm.N=newd.N=LReduceShort2(newd.a, newm.a, newd.N);
			//... before they enter new iteration
			DELETE_LIST(d); AUGMENT_LIST(d,newd);	//newd->d
			DELETE_LIST(m); AUGMENT_LIST(m,newm);	//newm->m
		}
	}

	DELETE_LIST(t);	//tags no longer needed
	DELETE_LIST(d);	//d no longer needed
	DELETE_LIST(m);	//m no longer needed

	//remove same primes from prl and prm
	prm.N=prl.N=LReduceShort2(prl.a, prm.a, prl.N);

	//count how many prime implicants each minterm is covered by
	DELETE_LIST(newm); W_L(newm, miLen-1, 0);	//newlist(dim(mi))->newm
	DELETE_LIST(newd);	//{}->newd
	//k counts newd
	for(i=0;i<miLen;i++) {
		for(j=0;j<prl.N;j++) {
			if((mi.a[i]|prm.a[j])==prl.a[j]) {
				PUSH_L(newd, j);
				W_L(newm, i, newm.a[i]+1);
			}
		}
	}
	//newm[]=number of primes for each minterm
	//newd[]=prime indexes for each minterm
	DELETE_LIST(mi); //mi no longer needed



	//Generating sums in the partitional polynomial - eg. (p1+p4+p5)*(p1+p2)...

	//ri[i]=index of the prime term at the i-th position
	short ri[miLen]; for(i=0;i<miLen;i++) ri[i]=0;
	//of[i]=pointer to the place in newd that belongs to the i-th minterm
	short of[miLen]; of[0]=0; for(i=1;i<miLen;i++) of[i]=newm.a[i-1]+of[i-1];
	//KK=number of all possible combinations for the partitional product
	long KK=1; for(i=0;i<miLen && KK<0x007fffffL;i++) KK*=newm.a[i];
	//if random solutions are being generated, user defines the number of iterations
	if(nRand) KK=nRand;
	//to prevent counter overflow and to avoid waiting 4194303 iterations
	if(KK>=0x007fffffL) {
		top_estack=PushString(top_estack, EExcessIter);
		DeleteAllLists();
		return;
	}
	//pprod[] = primes (their indexes) of individual partitional product eg. p1*p1*p3*p4
	CREATE_LIST(pprod,miLen,LQ);

	//collecting shortest pprod (each is initially miLen long, but get reduced, since p1*p1=p1 and so on)
	//pplist - shortest pprods of the same length get collected in it
	CREATE_LIST(pplist,0,LQ);	//{}->pplist
	short pplen=miLen;	//length of the currently shortest pprod

//+
	short nCovers=1;
//-
	long k;
	for(k=0;k<KK;k++) {	//iterate through all possible pprods
		//assemble one pprod
		if(nRand==0) {
			//iterative generation of pprods
			for(i=0;i<miLen;i++)
				pprod.a[i]=newd.a[of[i]+ri[i]];
			//counter
			for(i=0;i<miLen;i++) {
				ri[i]++;
				if(ri[i]<newm.a[i]) break;
				ri[i]=0;
			}
		} else {
			//random generation of pprods
			for(i=0;i<miLen;i++)
				pprod.a[i]=newd.a[of[i]+random(newm.a[i])];
		}

		//remove same products ( p1*p1=p1 )
		//lreduce(pprod)->pprod
		pprod.N=LReduceShort1(pprod.a, miLen);

		//Sort pprod for later removal of same pprods
		//lsorta(pprod)->pprod
		InsSortShort1(pprod.a, pprod.N);

		//collect shortest pprod to pplist
		if(pprod.N==pplen) {
			//if the same pprod follows, don't add it
			if(memcmp(&pplist.a[pplist.N-pplen], pprod.a, pplen*sizeof(short))!=0 || pplist.N==0) {
				AUGMENT_LIST(pplist,pprod);
			}
			nCovers=pplist.N/pplen;
		} else if(pprod.N<pplen) {
			DELETE_LIST(pplist);		//pprod->pplist
			AUGMENT_LIST(pplist,pprod);	//pprod->pplist
			pplen=pprod.N;
			nCovers=pplist.N/pplen;
		}
//+
		if(KK>127 && ((k&0x7f)==0)) {
			//displays:
			// k current combination : KK all combinations
			// P current minimal number of products in disjunction
			// c current number of covers of the same size found
			//   (may be lower after completion, as the same covers are then removed)
			sprintf(dbgstr,"k%li/%li P%i n%i  ",k,KK,pplen,nCovers);
			DrawStrXY(0,74,dbgstr,A_REPLACE);
			//read keys
			if(!OSdequeue(&key,kbd)) {
				if(key==KEY_ON) {
					//stop execution if 'ON' key is pressed
					push_longint(0);
					DeleteAllLists();
					return;
				} else if(key==KEY_ESC) {
					//if 'ESC' is pressed, return currently minimal cover
					break;
				}
			}
		}
//-
	}


	//remove same pprod from pplist
	//lreducen(pplist, pplen)->pplist
	pplist.N=LReduceShortSubL1(pplist.a, pplist.N, pplen);

	//seems like there should be no END_TAG in front of matrixes
	//it can be in front of lists, though, but it is not required
	top_estack--;
	//RETURN: rewrite primes numbers from pplist into matrix of primes and masks
	push_quantum(END_TAG);
	short primes[pplen],masks[pplen];
	for(i=0;i<pplist.N;i+=pplen) {
		for(j=0;j<pplen;j++) {
			masks[j]=prm.a[pplist.a[i+j]];
			primes[j]=prl.a[pplist.a[i+j]];
		}
		top_estack=PushShortIntList(top_estack, masks, pplen);
		top_estack=PushShortIntList(top_estack, primes, pplen);
	}
	push_quantum(LIST_TAG);

	DeleteAllLists();
}
