//Sizeable array
//Copyright (C) 2005 Ziga Mirai
//2005.07.07

//2006.04.06 comment translation

// intended for QMC
// includes safe memory allocation

#ifndef SZ_ARRAYH
#define SZ_ARRAYH

//---------------------------- Sizeable Array ------------------------------

#define MAX_N_LISTS 20

typedef struct {
	short *a;
	long N;
	long avail;
	long lq;
} TShortList;

//safety pointers to created lists
TShortList *pSLStruct[MAX_N_LISTS];
int nListsCreated;

//initializes array of pointers to the created lists
// must be included at the begining of the main() function!
#define USE_LISTS nListsCreated=0

//Releases memory of all created lists
void DeleteAllLists() {
	int i;
	for(i=0;i<nListsCreated;i++) {
		if(pSLStruct[i]->a!=NULL) {
			free(pSLStruct[i]->a);
			pSLStruct[i]->N=0;
			pSLStruct[i]->avail=0;
			pSLStruct[i]->a=NULL;
		}
	}
}

//A Not_enough_emory or too_many_lists_created safety return
void EMemoryExit() {
	DeleteAllLists();
	EatArgs();
	push_quantum(END_TAG);
	char rs[40];
	sprintf(rs,"E: Not enough emory (%02i)!",nListsCreated);
	top_estack=PushString(top_estack,rs);
	exit(0);
}

//CREATE_SA
// creates sizeable array structure named name
// filled with zeros
// you only call it once at the beginning
// it includes:
//  public a = array of the chosen type atype
//  public N = the length of the array (index to the topmost element)
//  prot. avail = available length of the array - available for direct writing
//  prot. lq = length quantum - length by which the array is extended when required
#define CREATE_LIST(name, len, len_quant)\
TShortList name={ NULL , len, len, len_quant};\
if(nListsCreated>=MAX_N_LISTS) EMemoryExit();\
pSLStruct[nListsCreated]=&name; nListsCreated++;\
if((len)>0) {\
	name.a=(short *) malloc(len*2);\
	if(name.a==NULL) EMemoryExit();\
	memset(name.a, 0, (len)*2);\
}

//DELETE_LIST
// reduces array length to 0 an frees memory
// can be called any time, but must be called at the end
#define DELETE_LIST(name) if(name.avail>0) free(name.a);\
name.N=name.avail=0; name.a=NULL;

//W_L
// inserts one element into a list
// increases N if necessary
// also extends the list if writing outside the reserved space
// extended portion is filled with zeros
#define W_L(name,i,e)\
if((i)>=name.avail) ExtendList(&name,i);\
name.a[i]=e;\
if((i)>=name.N) name.N=(i)+1

//PUSH_L
// pushes one element to the top of a list
#define PUSH_L(name,e) W_L(name,name.N,e);

//Extend_LIST
// it's private, user doesn't need this function
// reserves new block of memory an copies the content of the list when
//  writing outside its current length
void ExtendList(TShortList *list, long newLen) {
	long newavail=newLen+list->lq;
	if(list->a==NULL)
		list->a=malloc(newavail*2);
	else
		list->a=realloc(list->a, newavail*2);
	if(list->a==NULL) EMemoryExit();
	memset(list->a + list->avail, 0, (newavail - list->avail)*2 );
	list->avail=newavail;
}

//AUGMENT_LIST
//augments list name1 with contents of list name2
#define AUGMENT_LIST(name1,name2)\
if(name2.N>0) {\
	long oldN=name1.N;\
	W_L(name1, name1.N+name2.N-1,0);\
	memcpy(name1.a+oldN, name2.a, name2.N*2);\
}


//---------------------------^ Sizeable Array ^-----------------------------

#endif
